#!/usr/bin/env python
"""
This program exports SBML files from models given in metano's native format. The
exported models conform to SBML Level 2 Version 4, with the Flux Balance
Constraints extension: http://sbml.org/Documents/Specifications/SBML_Level_3/Packages/Flux_Balance_Constraints_(flux).


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

Alexander Krause contributed to this file.

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import absolute_import

from builtins import str
from builtins import zip
from builtins import range
from builtins import object
from metano.defines import Verbosity, COPYRIGHT_VERSION_STRING
from metano.sbmlstructure import SbmlDef
from metano.fba import OptionParser
from metano.reactionparser import ReactionParser
from metano.paramparser import ParamParser
from metano.metabolicmodel import Reactant, MetabolicModel
from metano.metabolicflux import MetabolicFlux
from numpy import inf
from copy import deepcopy
import xml.dom.minidom as dom
import os
import re


class SbmlExportStruc(object):
    SBML_URL = "http://www.sbml.org/sbml/level3/version1/core"
    XHTML_URL = "http://www.w3.org/1999/xhtml"
    MATH_URL = "http://www.w3.org/1998/Math/MathML"
    XMLNS_HTML = "xmlns:html"
    HTML_P = "html:p"
    ENCODING = "UTF-8"
    DEFAULT_COMPARTMENT = "Cytosol"
    BOUNDARY_COMPARTMENT = "Boundary"
    DEFAULT_BOUNDARY_SUFFIX = "_b"
    THE_FLUX_UNIT = "mmol_per_gDW_per_hr"
    UNITS = ({SbmlDef.KIND: "mole",   SbmlDef.EXPONENT: "1",  SbmlDef.SCALE: "-3", SbmlDef.MULTIPLIER: "1"},
             {SbmlDef.KIND: "gram",   SbmlDef.EXPONENT: "-1",
                 SbmlDef.SCALE: "1",  SbmlDef.MULTIPLIER: "1"},
             {SbmlDef.KIND: "second", SbmlDef.EXPONENT: "-1", SbmlDef.SCALE: "1",  SbmlDef.MULTIPLIER: "0.00027777"})


validIdCharactersPattern = re.compile("[\W_]+")


def removeInvalidIdCharacters(s):

    # remove all special chracters
    s = validIdCharactersPattern.sub("_", s)

    # append '_' to front if ID starts with a number
    s = s if s[0].isalpha() or s[0] == "_" else "_%s" % s

    return s


class SbmlWriter(object):
    """ Class for an SBML writer

    Member variables:

    modelName             -- name of SBML model to be written
    compartments          -- list of compartment IDs
    reactions             -- list of Reaction objects
    reactionLabels        -- list of reaction labels, indexed like reactions
    fluxes                -- flux values (MetabolicFlux object)
    metabolites           -- list of metabolite names
    metaboliteLabels      -- list of metabolite labels, indexed like metabolites
    metaboliteCompartment -- dict { metabolite : compartment }
    isBoundary            -- dict { metabolite : bool } - True for boundary met.
    infinityVal           -- finite value to be used instead of infinite bounds
                             (default: None)
    messages              -- error, warning, info, and debug messages generated
                             while exporting the model
    """

    _encodeBool = {False: "false", True: "true"}

    def __init__(self, model, modelName, maxmin, objStr="",
                 fluxes=MetabolicFlux(), stripComments=False, infinityVal=None):
        """ initialize class SMBLParser

        Keyword arguments:

        model              -- MetabolicModel to be exported to SBML
        modelName          -- name of SBML model to be written
        maxmin             -- True for maximization, False for minimization
        objStr             -- objective function definition (given as string)
        fluxes             -- flux values (MetabolicFlux object)
        infinityVal        -- finite value to be used instead of infinite bounds
                              (default: None)
        """
        self.clear()
        self.modelName = modelName

        useLabelsBak = model.useReactionLabels, model.useMetaboliteLabels
        model.useReactionLabels = True
        model.useMetaboliteLabels = True
        self.reactions = deepcopy(model.reactions)
        self.reactionLabels = deepcopy(model._getTranslatedReactionNames())
        self.reactionIndex = deepcopy(model.reactionDict)
        self.metabolites = deepcopy(model.metabolites)
        self.metaboliteLabels = deepcopy(model._getTranslatedMetaboliteNames())
        self.metaboliteIndex = deepcopy(model.metaboliteDict)
        model.useReactionLabels, model.useMetaboliteLabels = useLabelsBak

        self.isBoundary = dict((met, False) for met in self.metabolites)

        # Set coefficient vector of objective function
        try:
            # maxmin is inverted because in our SBML files (unlike in LP),
            # maximization is the default (i.e. factor 1 instead of -1)
            self.objCoef = ParamParser.convertObjFuncToLinVec(objStr,
                                                              model.reactionDict, maxmin=not maxmin)
        except ValueError:
            msg = ("Info: Objective function (%s) is not a linear expression. "
                   "%s will not be written." % (objStr, SbmlDef.P_OBJ_COEF))
            self.messages.append((Verbosity.INFO, msg))
        except KeyError as strerror:
            msg = ("Warning: An error occurred in evaluating the objective "
                   "function:\n %s." % strerror)
            self.messages.append((Verbosity.WARNING, msg))

        self.fluxes = fluxes
        self.stripComments = stripComments

        if infinityVal and infinityVal < 0.:
            infinityVal = -1.*infinityVal
        self.infinityVal = infinityVal

    def clear(self):
        """ clear all internal variables, warnings and info messages
        """
        self.modelName = ""
        self.compartments = []
        self.reactions = []
        self.reactionLabels = []
        self.fluxes = MetabolicFlux()
        self.objCoef = []
        self.metabolites = []
        self.metaboliteLabels = []
        self.metaboliteCompartment = {}
        self.isBoundary = {}
        self.messages = []

    def getMessages(self, level=Verbosity.INFO):
        """ return a list of all messages at or above the given level of
            severity

        Levels are defined in defines.py.
        """
        return [x[1] for x in self.messages if x[0] <= level]

    def makeCompartments(self, compartmentRegex="",
                         defaultCompartment=SbmlExportStruc.DEFAULT_COMPARTMENT):
        """ construct compartment IDs and metabolite->compartment mapping from
            self.metabolites

        Keyword arguments:

        compartmentRegex   -- regular expression describing what part (if any)
                              of a metabolite name identifies the compartment -
                              must contain exactly one group, which is used to
                              generate compartment IDs
        defaultCompartment -- ID of compartment of metabolites not matching
                              compartmentRegex

        Modified member variables:

        compartments, metaboliteCompartment

        Returns True if successful, else False
        """
        if not compartmentRegex:
            # Assign all metabolites to the same (default) compartment
            self.compartments = [defaultCompartment]
            self.metaboliteCompartment = dict((met, defaultCompartment)
                                              for met in self.metabolites)
            msg = ("Info: No compartment reg-ex given. Assigned all metabolites"
                   " to default compartment (%s)." % defaultCompartment)
            self.messages.append((Verbosity.INFO, msg))
            return True

        try:
            pattern = re.compile(compartmentRegex)
        except re.error as strerror:
            msg = ("Error in regular expression: %s." %
                   strerror)
            self.messages.append((Verbosity.ERROR, msg))
            return False

        setOfCompartments = set()

        for met in self.metabolites:
            reHit = pattern.search(met)
            if not reHit:
                # Assign non-matching metabolites to default compartment
                compartment = defaultCompartment
                msg = ("Debug: Metabolite '%s' does not match compartment "
                       "reg-ex. Assigned to default compartment (%s)." %
                       (met, defaultCompartment))
                self.messages.append((Verbosity.DEBUG, msg))
            else:
                try:
                    compartment = reHit.group(1)
                except IndexError:
                    msg = ("Error: Regular expression has no groups. Use "
                           "parentheses to define one.\n"
                           "       Example: \"_(ex|c|p)$\".")
                    self.messages.append((Verbosity.ERROR, msg))
                    return False

            compartment = removeInvalidIdCharacters(compartment)
            self.metaboliteCompartment[met] = compartment

            if compartment not in setOfCompartments:
                self.compartments.append(compartment)
                setOfCompartments.add(compartment)
                msg = "Info: Created compartment '%s'." % compartment
                self.messages.append((Verbosity.INFO, msg))

        return True

    def addBoundaryMetabolites(self,
                               suffix=SbmlExportStruc.DEFAULT_BOUNDARY_SUFFIX):
        """ add boundary metabolites to all reactions that have an empty left-
            or right-hand side

        Boundary metabolites are marked with the given suffix and assigned their
        own compartment ("Boundary"). No non-boundary metabolite may have the
        designated boundary suffix.

        Returns True if successful, else False
        """
        for metabolite in self.metabolites:
            if metabolite.endswith(suffix):
                msg = ("Error: Non-boundary metabolite '%s' found with "
                       "designated boundary suffix '%s'. Please choose a unique"
                       " suffix." % (metabolite, suffix))
                self.messages.append((Verbosity.ERROR, msg))
                return False

        hasBoundaryCompartment = False
        incompleteReactions = [r for r in self.reactions if not r.getEducts() or
                               not r.getProducts()]

        for reaction in incompleteReactions:
            educts = reaction.getEducts()
            products = reaction.getProducts()

            if not educts and not products:
                # Empty reaction, illegal in SBML (also in metano)
                i = self.reactionIndex[reaction.name]
                del self.reactions[i]
                del self.reactionLabels[i]
                # Rebuild index
                self.reactionIndex = dict(list(zip([r.name for r in self.reactions],
                                                   list(range(len(self.reactions))))))
                msg = ("Warning: Reaction '%s' has neither reactants "
                       "nor products. Deleted." % reaction.name)
                self.messages.append((Verbosity.WARNING, msg))

            elif products == []:
                if not hasBoundaryCompartment:
                    hasBoundaryCompartment = self._addBoundaryCompartment()

                self._addBoundaryMetabolitesToSide(reaction, "right", educts,
                                                   suffix)
            elif educts == []:
                if not hasBoundaryCompartment:
                    hasBoundaryCompartment = self._addBoundaryCompartment()

                self._addBoundaryMetabolitesToSide(reaction, "left", products,
                                                   suffix)
        return True

    def _addBoundaryCompartment(self):
        """ add "Boundary" compartment to list of compartments (if not already
            present)

            Returns True
        """
        if SbmlExportStruc.BOUNDARY_COMPARTMENT not in self.compartments:
            self.compartments.append(SbmlExportStruc.BOUNDARY_COMPARTMENT)
        return True

    def _addBoundaryMetabolitesToSide(self, reaction, side, reactants, suffix):
        """ add boundary metabolites to the specified side of the given reaction

        Keyword arguments:

        reaction  -- Reaction object to be modified
        side      -- "left" or "right"
        reactants -- list of Reactant objects for the other (non-boundary) side
        suffix    -- suffix for marking the newly created metabolites
        """
        for reactant in reactants:
            boundaryName = reactant.name + suffix
            boundaryReactant = Reactant(boundaryName, -1.*reactant.coef)

            if boundaryName not in self.metaboliteCompartment:
                self.metabolites.append(boundaryName)
                self.metaboliteCompartment[boundaryName] = \
                    SbmlExportStruc.BOUNDARY_COMPARTMENT
                self.isBoundary[boundaryName] = True
                orig_index = self.metaboliteIndex[reactant.name]
                boundaryLabel = (self.metaboliteLabels[orig_index] + suffix)
                self.metaboliteLabels.append(boundaryLabel)
                self.metaboliteIndex[boundaryName] = len(self.metaboliteIndex)

            reaction.addReactant(boundaryReactant)
            msg = ("Debug: Added boundary species '%s' with "
                   "coefficient %r to %s-hand side of reaction '%s'."
                   % (boundaryName, abs(reactant.coef), side, reaction.name))
            self.messages.append((Verbosity.DEBUG, msg))

    def writeFile(self, sbmlFileName):
        """ write prepared structures to file designated by sbmlFileName
        """
        sbmlTree = dom.Document()

        sbml = dom.Element(SbmlDef.SBML)
        sbml.setAttribute(SbmlDef.LEVEL, "3")
        sbml.setAttribute(SbmlDef.VERSION, "1")
        sbml.setAttribute(SbmlDef.XMLNS, SbmlExportStruc.SBML_URL)
        sbml.setAttribute(SbmlExportStruc.XMLNS_HTML,
                          SbmlExportStruc.XHTML_URL)

        model = dom.Element(SbmlDef.MODEL)
        model.setAttribute(
            SbmlDef.ID, removeInvalidIdCharacters(self.modelName))
        model.setAttribute(SbmlDef.NAME, self.modelName)

        # Write unit definition
        listOfUnitDefinitions = dom.Element(SbmlDef.LISTOFUNITDEFINITIONS)
        unitDefinition = dom.Element(SbmlDef.UNITDEFINITION)
        unitDefinition.setAttribute(
            SbmlDef.ID, removeInvalidIdCharacters(SbmlExportStruc.THE_FLUX_UNIT))
        listOfUnits = dom.Element(SbmlDef.LISTOFUNITS)
        for u in SbmlExportStruc.UNITS:
            unit = dom.Element(SbmlDef.UNIT)
            for attr in u:
                unit.setAttribute(attr, u[attr])
            listOfUnits.appendChild(unit)
        unitDefinition.appendChild(listOfUnits)
        listOfUnitDefinitions.appendChild(unitDefinition)
        model.appendChild(listOfUnitDefinitions)

        # Write compartments
        listOfCompartments = dom.Element(SbmlDef.LISTOFCOMPARTMENTS)
        for compartment in self.compartments:
            element = dom.Element(SbmlDef.COMPARTMENT)
            element.setAttribute(
                SbmlDef.ID, removeInvalidIdCharacters(compartment))
            element.setAttribute(SbmlDef.CONSTANT, "true")
            listOfCompartments.appendChild(element)
        model.appendChild(listOfCompartments)

        # Create list of species
        listOfSpecies = dom.Element(SbmlDef.LISTOFSPECIES)
        for i in range(len(self.metabolites)):
            metabolite = self.metabolites[i]
            metaboliteLabel = self.metaboliteLabels[i]
            metaboliteCompartment = self.metaboliteCompartment[metabolite]
            boundaryCondition = self._encodeBool[self.isBoundary[metabolite]]
            species = dom.Element(SbmlDef.SPECIES)
            species.setAttribute(
                SbmlDef.ID, removeInvalidIdCharacters(metabolite))
            species.setAttribute(SbmlDef.NAME, metaboliteLabel)
            species.setAttribute(SbmlDef.COMPARTMENT, metaboliteCompartment)
            species.setAttribute(SbmlDef.BOUNDARYCONDITION, boundaryCondition)
            species.setAttribute(SbmlDef.HASONLYSUBSTANCEUNITS, "true")
            species.setAttribute(SbmlDef.CONSTANT, "false")
            listOfSpecies.appendChild(species)

            msg = "Debug: Created species node '%s' " % metabolite
            if metaboliteLabel and metaboliteLabel != metabolite:
                msg += "(%s) " % metaboliteLabel
            msg += ("with %s='%s' and %s='%s'." % (SbmlDef.COMPARTMENT,
                                                   metaboliteCompartment, SbmlDef.BOUNDARYCONDITION,
                                                   boundaryCondition))
            self.messages.append((Verbosity.DEBUG, msg))

        model.appendChild(listOfSpecies)

        # Create list of reactions
        listOfReactions = self._createListOfReactions()
        model.appendChild(listOfReactions)

        sbml.appendChild(model)
        sbmlTree.appendChild(sbml)

        # Write tree to file
        with open(sbmlFileName, 'w') as f:
            sbmlTree.writexml(f, "", '\t', '\n', SbmlExportStruc.ENCODING)

    def _createListOfReactions(self):
        listOfReactions = dom.Element(SbmlDef.LISTOFREACTIONS)
        fluxVec = self.fluxes.getVecOrderedByDict(self.reactionIndex)
        hasFluxes = len([f for f in fluxVec if f is not None]) != 0
        if self.fluxes and not hasFluxes:
            # Generate warning if model and flux distribution do not share any
            # reaction (else: generate a warning for each reaction without flux)
            msg = ("Warning: The given flux distribution does not fit the model"
                   " (reaction names are different).")
            self.messages.append((Verbosity.WARNING, msg))

        for i in range(len(self.reactions)):
            # Set required attributes for reaction
            rea_item = self.reactions[i]
            reactionName = rea_item.name
            reactionLabel = self.reactionLabels[i]
            reversible = self._encodeBool[rea_item.direction == 0]
            reaction = dom.Element(SbmlDef.REACTION)
            reaction.setAttribute(
                SbmlDef.ID, removeInvalidIdCharacters(reactionName))
            reaction.setAttribute(SbmlDef.NAME, reactionLabel)
            reaction.setAttribute(SbmlDef.REVERSIBLE, reversible)
            reaction.setAttribute(SbmlDef.FAST, "false")

            msg = "Debug: Created reaction node '%s' " % reactionName
            if reactionLabel and reactionLabel != reactionName:
                msg += "(%s) " % reactionLabel
            msg += "with %s='%s'." % (SbmlDef.REVERSIBLE, reversible)
            self.messages.append((Verbosity.DEBUG, msg))

            if rea_item.direction == -1:
                # Irreversible reaction proceeding from right to left must be
                # reversed - in SBML, irreversible implies left to right
                educts, products = rea_item.getProducts(), rea_item.getEducts()
                tmp_ub, tmp_lb = -rea_item._lb, -rea_item._ub
                tmp_flux = -fluxVec[i] if fluxVec[i] is not None else None
                msg = ("Debug: Replacing irreversible right-to-left reaction "
                       "'%s' with left-to-right reaction." % reactionName)
                self.messages.append((Verbosity.DEBUG, msg))
            else:
                educts, products = rea_item.getEducts(), rea_item.getProducts()
                tmp_lb, tmp_ub = rea_item.getBounds()
                tmp_flux = fluxVec[i]

            # Create notes
            if not self.stripComments:
                comments = str(rea_item.comment).splitlines()
                if len(comments) > 0:
                    notes = dom.Element(SbmlDef.NOTES)
                    for line in comments:
                        paragraph = dom.Element(SbmlExportStruc.HTML_P)
                        t = dom.Text()
                        t.data = line.strip()
                        paragraph.appendChild(t)
                        notes.appendChild(paragraph)
                    reaction.appendChild(notes)

            # Create list of reactants and list of products
            if educts:
                listOfReactants = self._createListOfReactantsProducts(
                    SbmlDef.LISTOFREACTANTS, educts)
                reaction.appendChild(listOfReactants)
            if products:
                listOfProducts = self._createListOfReactantsProducts(
                    SbmlDef.LISTOFPRODUCTS, products)
                reaction.appendChild(listOfProducts)

            # Create kineticLaw for reaction
            kineticLaw = dom.Element(SbmlDef.KINETICLAW)

            math = dom.Element(SbmlDef.MATH)
            math.setAttribute(SbmlDef.XMLNS, SbmlExportStruc.MATH_URL)
            # ===================================================================
            # ci = dom.Element(SbmlDef.CI)
            # t = dom.Text()
            # t.data = "%s" % SbmlDef.P_FLUX_VALUE
            # ci.appendChild(t)
            # math.appendChild(ci)
            # ===================================================================
            cn = dom.Element(SbmlDef.CN)
            t = dom.Text()

            # Write listOfParameters with lower bound, upper bound, objective
            # coefficient, and flux value
            listOfParameters = dom.Element(SbmlDef.LISTOFPARAMETERS)
            paramMsg = "Debug: Parameters: "

            if self.infinityVal or tmp_lb != -inf:

                if tmp_lb != -inf:
                    lb = tmp_lb
                else:
                    lb = -1*self.infinityVal

                parameter = dom.Element(SbmlDef.PARAMETER)
                parameter.setAttribute(
                    SbmlDef.ID, removeInvalidIdCharacters(SbmlDef.P_LOWER_BOUND))
                # parameter.setAttribute(SbmlDef.CONSTANT, "true") # Change in SBML V3
                value = repr(lb)
                parameter.setAttribute(SbmlDef.VALUE, value)
                parameter.setAttribute(SbmlDef.UNITS,
                                       SbmlExportStruc.THE_FLUX_UNIT)
                listOfParameters.appendChild(parameter)
                paramMsg += "%s: %r" % (SbmlDef.P_LOWER_BOUND, lb)
            else:
                paramMsg += "no lower bound"

            if self.infinityVal or tmp_ub != inf:

                if tmp_ub != inf:
                    ub = tmp_ub
                else:
                    ub = self.infinityVal

                parameter = dom.Element(SbmlDef.PARAMETER)
                parameter.setAttribute(
                    SbmlDef.ID, removeInvalidIdCharacters(SbmlDef.P_UPPER_BOUND))
                # parameter.setAttribute(SbmlDef.CONSTANT, "true") # Change in SBML V3
                value = repr(ub)
                parameter.setAttribute(SbmlDef.VALUE, value)
                parameter.setAttribute(SbmlDef.UNITS,
                                       SbmlExportStruc.THE_FLUX_UNIT)
                listOfParameters.appendChild(parameter)
                paramMsg += ", %s: %r" % (SbmlDef.P_UPPER_BOUND, ub)
            else:
                paramMsg += ", no upper bound"

            if self.objCoef:
                coef = self.objCoef[i]
                parameter = dom.Element(SbmlDef.PARAMETER)
                parameter.setAttribute(
                    SbmlDef.ID, removeInvalidIdCharacters(SbmlDef.P_OBJ_COEF))
                # parameter.setAttribute(SbmlDef.CONSTANT, "true") # Change in SBML V3
                value = repr(coef)
                parameter.setAttribute(SbmlDef.VALUE, value)
                listOfParameters.appendChild(parameter)
                paramMsg += ", %s: %r" % (SbmlDef.P_OBJ_COEF, coef)

            if tmp_flux is not None:
                parameter = dom.Element(SbmlDef.PARAMETER)
                parameter.setAttribute(
                    SbmlDef.ID, removeInvalidIdCharacters(SbmlDef.P_FLUX_VALUE))
                # parameter.setAttribute(SbmlDef.CONSTANT, "true") # Change in SBML V3
                value = repr(tmp_flux)
                parameter.setAttribute(SbmlDef.VALUE, value)
                parameter.setAttribute(SbmlDef.UNITS,
                                       SbmlExportStruc.THE_FLUX_UNIT)
                listOfParameters.appendChild(parameter)
                paramMsg += ", %s: %r" % (SbmlDef.P_FLUX_VALUE, tmp_flux)
            elif hasFluxes:
                msg = ("Warning: Reaction '%s' has no flux value." %
                       reactionName)
                self.messages.append((Verbosity.WARNING, msg))

            t.data = "%s" % value
            cn.appendChild(t)
            math.appendChild(cn)
            kineticLaw.appendChild(math)

            kineticLaw.appendChild(listOfParameters)
            reaction.appendChild(kineticLaw)
            self.messages.append((Verbosity.DEBUG, paramMsg))

            # Store reaction in the reaction list
            listOfReactions.appendChild(reaction)

        return listOfReactions

    def _createListOfReactantsProducts(self, typ, reactants):
        """ create listOfReactants or listOfProducts node

        Keyword arguments:

        typ       -- SbmlDef.LISTOFREACTANTS or SbmlDef.LISTOFPRODUCTS
        reactants -- list of Reactant objects

        Returns SBML node with the given type
        """
        listOfSpeciesRef = dom.Element(typ)
        for r in reactants:
            reactant = dom.Element(SbmlDef.SPECIESREFERENCE)
            reactant.setAttribute(
                SbmlDef.SPECIES, removeInvalidIdCharacters(r.name))
            reactant.setAttribute(SbmlDef.CONSTANT, "false")
            reactant.setAttribute(SbmlDef.STOICHIOMETRY, repr(abs(r.coef)))
            listOfSpeciesRef.appendChild(reactant)

        return listOfSpeciesRef


def main():
    # 1. Parse command line

    usage = "Usage: %prog [options]"
    version = "SBML writer\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)

    parser.add_option("-r", "--reactions", dest="reactionFile",
                      help="write an SBML file for the network given "
                           "by the reaction FILE", metavar="FILE")
    parser.add_option("-p", "--parameters", dest="paramFile",
                      help="use the given scenario FILE for SBML export",
                      metavar="FILE")
    parser.add_option("-q", "--flux", dest="fluxFile", help="write fluxes read "
                      "from FILE to SBML file (optional)", metavar="FILE")
    parser.add_option("-o", "--output", dest="outputFile",
                      help="write output to SBML FILE",
                      metavar="FILE")
    parser.add_option("-c", "--compartment-regex", dest="compartmentRegex",
                      help="regular expression PATTERN in metabolite names that"
                      " identifies compartment information (e.g. "
                      "\"\\[(\\w+)\\]$\" or \"_(ex|c|p)$\") -- must contain "
                      "exactly one group (in parentheses)", metavar="PATTERN")
    parser.add_option("-d", "--default-compartment", dest="defaultCompartment",
                      help="NAME of default compartment (assigned to all "
                      "metabolites not matching the compartment reg-ex, "
                      "default: Cytosol)", metavar="NAME")
    parser.add_option("-s", "--suppress-comments", dest="noComments",
                      help="do not transfer comments to SBML file",
                      action="store_true")
    parser.add_option("-i", "--infinity", dest="infinityVal", type="float",
                      help="use finite VALUE instead of infinity (defaults to "
                      "None)", metavar="VALUE")
    parser.add_option("-e", "--debug-info", dest="debugInfo", help="print"
                      " debug information", action="store_true")
    parser.add_option("-b", "--boundary-suffix", dest="boundarySuffix",
                      help="add boundary metabolites, marked with the given "
                      "SUFFIX to reactions lacking either products or "
                      "reactants", metavar="SUFFIX")

    parser.set_defaults(defaultCompartment=SbmlExportStruc.DEFAULT_COMPARTMENT,
                        noComments=False, debugInfo=False)

    options, _ = parser.parse_args()
    parser.check_required("-r")
    parser.check_required("-p")
    parser.check_required("-o")

    print("Start creating SBML file '%s'.." % options.outputFile)
    print("Reaction file:           '%s'" % options.reactionFile)
    print("Scenario file:           '%s'" % options.paramFile)
    if options.fluxFile:
        print("Flux file:               '%s'" % options.fluxFile)
    print("Compartment reg-ex:      '%s'" % options.compartmentRegex)
    print("Default compartment:     '%s'" % options.defaultCompartment)
    print("Infinity value:          %r" % options.infinityVal)
    if options.boundarySuffix:
        print("Boundary suffix:         '%s'\n" % options.boundarySuffix)

    # 2. Parse reaction file

    rparser = ReactionParser()
    model = MetabolicModel()
    try:
        model.addReactionsFromFile(options.reactionFile, rparser)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in reaction file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()

    # 3. Parse scenario file

    model_messages = []
    pparser = ParamParser()
    try:
        # Read objective function and bounds from scenario file
        maxmin, objStr, _, _, lb, ub = pparser.parse(options.paramFile)
        # Set flux bounds in model
        model.setFiniteBounds(lb, ub, True, model_messages)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.paramFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in scenario file %s:" %
               os.path.basename(options.paramFile))
        print(strerror)
        exit()
    except ValueError as strerror:
        print(strerror)
        exit()

    # Show warning and info messages of parsers
    msgs = (rparser.getMessages() + pparser.getMessages() +
            [x[1] for x in model_messages])
    if msgs:
        print('\n'.join(msgs))

    # 4. Read flux distribution from file

    fluxes = MetabolicFlux()
    if options.fluxFile:
        try:
            fluxes.readFromFile(options.fluxFile)
        except IOError as strerror:
            print ("An error occurred while trying to read file %s:" %
                   os.path.basename(options.fluxFile))
            print(strerror)
            print("Fluxes will not be written.")
            fluxes = MetabolicFlux()
        except SyntaxError as strerror:
            print ("An error occurred parsing file %s:" %
                   os.path.basename(options.fluxFile))
            print(strerror)
            print("Fluxes will not be written.")
            fluxes = MetabolicFlux()

    # 5. Write SBML file

    # Name of model is name of output file without file extension
    modelName = os.path.basename(options.outputFile)
    dotPos = modelName.rfind(".")
    if dotPos > 0:
        modelName = modelName[:dotPos]

    sbmlWriter = SbmlWriter(model, modelName, maxmin, objStr, fluxes,
                            options.noComments, options.infinityVal)
    success = sbmlWriter.makeCompartments(options.compartmentRegex,
                                          options.defaultCompartment)
    if success:
        if options.boundarySuffix:
            success = sbmlWriter.addBoundaryMetabolites(options.boundarySuffix)
        if success:
            sbmlWriter.writeFile(options.outputFile)

    if options.debugInfo:
        msgs = sbmlWriter.getMessages(Verbosity.DEBUG)
    else:
        # Show only error, warning, and info messages
        msgs = sbmlWriter.getMessages()
    if msgs:
        print('\n'+'\n'.join(msgs))

    print("Finished.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

""" create scatterplot from scenario file

This script will create a scatterplot from given flux distributions or
scenario files.

It parses a reaction file and a scenario file, formulates the
corresponding linear, quadratic, or other nonlinear optimization problem,
and calls an appropriate solver.


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import division

from builtins import zip
from builtins import range
from builtins import object
from past.utils import old_div
import os
import random
import re
import argparse
import configparser
import matplotlib.pyplot as plt
from metano.metabolicflux import MetabolicFlux
from metano.fba import FbAnalyzer
from metano.fva import FvAnalyzer
from metano.mfm import MfmAnalyzer
from metano.metabolicmodel import MetabolicModel
from metano.reactionparser import ReactionParser
from metano.paramparser import ParamParser
from metano.defines import FbaParam, COPYRIGHT_VERSION_STRING

try:
    from adjustText import adjust_text
    ADJ_TEXT_AV = True
except ImportError:
    ADJ_TEXT_AV = False
    print("""Warning: The 'adjustText' package is not installed.
             It is highly recommended to use adjustText when annotations are shown,
             since the annotation placement is significantly improved.\n
             To install adjustText, please type:\n
             (sudo) pip install adjustText""")

ALLOWED_METHODS = ["fba", "fva", "mfm", "mva", "moma"]


def abs_diff(a_vec, b_vec):
    """ calculate the absolute difference between two vectors """
    if max(a_vec) > max(b_vec):
        return min(a_vec) - max(b_vec)
    return min(b_vec) - max(a_vec)


def abbreviate_all_in_list(str_list, max_len=20, ids=False, unique=False):
    """ abbreviate all strings in given list """
    result = []
    letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    special_chars = ('_', '-', '"', "'", ',')
    lower_vowels = ('a', 'e', 'i', 'o', 'u')
    upper_vowels = ('A', 'E', 'I', 'O', 'U')

    for item in str_list:
        found = False
        if ids:
            # use database ids if available
            regex_id = {"brenda": re.compile(r"(B[S|R]\d+)"),
                        "metacyc": re.compile(r"(RXN[A-Z\-\d]+|[A-Z\.\-\d\+]+-RXN)"),
                        "kegg": re.compile(r"(R\d+)")}

            for database in regex_id:
                if regex_id[database].findall(item):
                    found = True
                    result.append(regex_id[database].findall(item)[0])
                    break
        # if no id found or id search not applicable
        if not found:
            item = item.strip()
            if len(item) <= max_len:
                if item in result:
                    item = item[:-1] + random.choice(letters)
                else:
                    result.append(item)
                    continue
            while len(item) > max_len:
                # always start to remove items from the end of the string
                # first delete all lower vowels
                if any(l in item for l in lower_vowels):
                    index = max([item.rfind(l) for l in lower_vowels])
                    item = item[:index] + item[index+1:]
                # then delete all upper vowels
                elif any(l in item for l in upper_vowels):
                    index = max([item.rfind(l) for l in upper_vowels])
                    item = item[:index] + item[index+1:]
                # next delete all special characters
                elif any(l in item for l in special_chars):
                    index = max([item.rfind(l) for l in special_chars])
                    item = item[:index] + item[index+1:]
                else:
                    # at the end cut the remaining elements from max_len
                    item = item[0:max_len]
            # if each item should be unique in list:
            if unique:
                j = 0
                while item in result:
                    j += 1
                    item = item[:-1] + random.choice(letters)
                    if j > 20:
                        break
            result.append(item)
    return result


class Scatterplot(object):
    """ Class for the generation of a scatterplot"""

    def __init__(self, method, filenames, model=None):
        super(Scatterplot, self).__init__()

        self.method = method.lower()
        if self.method not in ALLOWED_METHODS:
            print ("Error: Method %r is unknown. Please choose one of the following "
                   "methods: FBA, FVA, MFM, MVA (only for sce)" % (self.method))
            exit()

        self.filenames = filenames
        if len(self.filenames) != 2:
            print("Error: -f option needs exactly 2 solution or "
                  "scenario files to create a scatterplot.")
            exit()

        self.model = model
        self.result = {}

        self.style = {
            "label_length": "15",
            "ids_in_names": "False",
            "unique_names": "False",
            "min_value": "0",
            "insignificant_color": "#545454",
            "significant_color": "#BE1E3C",  # TU BS red
            "xlabel": self.filenames[0],
            "ylabel": self.filenames[1],
            "title": "",
            "show_errorbar": "significant",  # significant, all, none
            "show_labels": "significant",  # significant, all, none
            "arrowcolor": "#BE1E3C",
            "labelcolor": "#BE1E3C",
            "significance_threshold": "0.1",
            "fontsize": "10",
            "figsize": "10 8",
            "plot_grid": "False",
            "arrow_line_width": "1",
        }

        if self.is_sce():
            if not self.model:
                print("Error: a reaction file is needed to perform "
                      "analysis from given scenario files!")
                exit()
            self.do_analysis()
        else:
            self.read_solution()

    def read_solution(self):
        """ read flux distribution/MFM from file """
        for solution in self.filenames:
            if self.method in ["fba", "moma"]:
                flux = MetabolicFlux()
                try:
                    flux.readFromFile(solution)
                except IOError as strerror:
                    print ("An error occurred while trying to read file %s:" %
                           os.path.basename(solution))
                    print(strerror)
                    exit()
                except SyntaxError as strerror:
                    print ("An error occurred parsing file %s:" %
                           os.path.basename(solution))
                    print(strerror)
                    exit()

                for rea in flux.fluxDict:
                    self.result[rea] = self.result[rea] + [[flux.fluxDict[rea]]] \
                        if rea in self.result else [[flux.fluxDict[rea]]]

            elif self.method == "fva":
                try:
                    flux, minmax = \
                        FvAnalyzer.parseSolutionFile(solution)
                except IOError as strerror:
                    print ("An error occurred while trying to read file %s:" %
                           os.path.basename(solution))
                    print(strerror)
                    exit()
                except SyntaxError as strerror:
                    print ("Error in FVA solution file %s:" %
                           os.path.basename(solution))
                    print(strerror)
                    exit()

                for rea in flux.fluxDict:
                    self.result[rea] = self.result[rea] + [[flux[rea]]+list(minmax[rea])] \
                        if rea in self.result else [[flux[rea]]+list(minmax[rea])]

            elif self.method == "mfm":
                solution = MfmAnalyzer.parseSolutionFile(solution)
                for met in solution:
                    self.result[met] = self.result[met] + [[solution[met]]] \
                        if met in self.result else [[solution[met]]]

            else:
                print ("Error: only FBA/MOMA, FVA, and MFM solutions can be read from file.\nIf "
                       "you want to calculate MVA, please provide two scenario files and a model!")
                exit()

    def is_sce(self):
        """ check if the given files are parameter files """
        for filename in self.filenames:
            try:
                # Parse file, get maxmin, name of objective function, and solver name
                pparser = ParamParser()
                pparser.parse(filename)
            except Exception:
                return False
        return True

    def do_analysis(self):
        """ perform FBA/FVA/MFM/MVA depending on given method """
        tolerance = 0.95

        for filename in self.filenames:
            model = MetabolicModel()
            rparser = ReactionParser()
            try:
                model.addReactionsFromFile(self.model, rparser)
            except IOError as strerror:
                print ("An error occurred while trying to read file %s:" %
                       os.path.basename(self.model))
                print(strerror)
                exit()
            except SyntaxError as strerror:
                print ("Error in reaction file %s:" %
                       os.path.basename(self.model))
                print(strerror)
                exit()
            # model = self.model
            model_messages = []
            pparser = ParamParser()
            # Parse file, get maxmin, name of objective function, and solver name
            maxmin, obj_str, solver, num_iter, lb_vec, ub_vec = pparser.parse(
                filename)
            fba_params = FbaParam(solver, maxmin, obj_str, num_iter)
            fba_params.setLinConstraints(pparser.lin_constraints)
            print(filename)
            # Set flux bounds in model
            model.setFiniteBounds(lb_vec, ub_vec, True, model_messages)

            solver = "default" if not solver else solver

            if self.method in ["fba", "moma"]:
                fba = FbAnalyzer(solver)
                _, solution, _ = fba.runOnModel(
                    model, fba_params, True, False, True)
                for rea in model.reactionDict:
                    self.result[rea] = self.result[rea] + [[solution[rea]]] \
                        if rea in self.result else [[solution[rea]]]

            elif self.method == "fva":
                fva = FvAnalyzer(solver)
                minmax, solution, _, _, _ = fva.runOnModel(
                    model, fba_params, tolerance)

                if not minmax:
                    print(
                        "FVA Warning: Model is infeasible or unbounded. Nothing to do.")
                    exit()
                for rea in model.reactionDict:
                    i = model.reactionDict[rea]
                    self.result[rea] = self.result[rea] + [[solution[i]]+list(minmax[i])] \
                        if rea in self.result else [[solution[i]]+list(minmax[i])]

            elif self.method == "mfm":
                mfm = MfmAnalyzer(solver)
                solution, _ = mfm.runOnModel(
                    model, fba_params, tolerance, minimize=True)
                for met in model.metaboliteDict:
                    i = model.metaboliteDict[met]
                    self.result[met] = self.result[met] + [[solution[i]]] \
                        if met in self.result else [[solution[i]]]

            elif self.method == "mva":
                fba = FbAnalyzer(solver)
                _, solution, _ = fba.runOnModel(
                    model, fba_params, True, False, True)

                mfm = MfmAnalyzer(solver)
                solution_min, _ = mfm.runOnModel(
                    model, fba_params, tolerance, minimize=True)
                solution_max, _ = mfm.runOnModel(
                    model, fba_params, tolerance, minimize=False)

                split_ratios = solution.computeAllSplitRatios(
                    model, 0., False, False)

                for met in model.metaboliteDict:
                    i = model.metaboliteDict[met]
                    flux = 0
                    for rea in split_ratios[met][0]:
                        flux += split_ratios[met][0][rea][1]

                    self.result[met] = self.result[met] + \
                        [[flux, solution_min[i], solution_max[i]]]\
                        if met in self.result else \
                        [[flux, solution_min[i], solution_max[i]]]

            else:
                print("No valid method was chosen.")
                exit()

    def set_plot_style(self, stylefile=None):
        """ read style for the scatterplot
            from configfile or take defaults """
        # first set defaults:
        config = configparser.SafeConfigParser(self.style)

        # try to read configfile, if exists
        if stylefile:
            try:
                open(stylefile)
            except IOError as error:
                print(error)
                print("Please enter a valid config file.")
                exit()
            config.read(stylefile)

        # parse style in a dictionary
        try:
            self.style = {
                "label_length": int(config.get("DEFAULT", "label_length")),
                "ids_in_names": config.get("DEFAULT", "ids_in_names") == "True",
                "unique_names": config.get("DEFAULT", "unique_names") == "True",
                "min_value": float(config.get("DEFAULT", "min_value")),
                "insignificant_color": config.get("DEFAULT", "insignificant_color"),
                "significant_color": config.get("DEFAULT", "significant_color"),
                "xlabel": config.get("DEFAULT", "xlabel"),
                "ylabel": config.get("DEFAULT", "ylabel"),
                "title": config.get("DEFAULT", "title"),
                "show_errorbar": config.get("DEFAULT", "show_errorbar"),
                "show_labels": config.get("DEFAULT", "show_labels"),
                "arrowcolor": config.get("DEFAULT", "arrowcolor"),
                "labelcolor": config.get("DEFAULT", "labelcolor"),
                "significance_threshold": float(config.get("DEFAULT", "significance_threshold")),
                "fontsize": float(config.get("DEFAULT", "fontsize")),
                "figsize": tuple(int(i) for i in config.get("DEFAULT", "figsize").split()),
                "plot_grid": config.get("DEFAULT", "plot_grid") == "True",
                "arrow_line_width": float(config.get("DEFAULT", "arrow_line_width")),
            }
        except Exception as error:
            print(error)
            print("Please check the configuration file")
            exit()

    def do_scatterplot(self, outputfile=None, obj_fun=None, stylefile=None):
        """ draw scatterplot """
        # read stylefile or get default style:
        self.set_plot_style(stylefile)
        style = self.style
        # calculate obj jun coefficient if obj fun is given
        if obj_fun:
            try:
                obj_coeff = old_div(self.result[obj_fun][0][0],
                                    self.result[obj_fun][1][0])
            except KeyError:
                print ("Error: Objective function %r could not be found in reactions dict!"
                       % obj_fun)
                exit()
        else:
            obj_coeff = 1

        # remove all entries that have only one result or are zero at some point
        self.result = {i: self.result[i] for i in self.result if len(self.result[i]) == 2
                       and self.result[i][0][0] > style["min_value"]
                       and self.result[i][1][0] > style["min_value"]}

        # create canvas for drawing the scatterplot
        plt.figure(figsize=style["figsize"])

        # parse result dict
        if self.method in ["fva", "mva"]:
            name, x, xmin, xmax, y, ymin, ymax = list(zip(*[
                ([rea] + self.result[rea][0] + self.result[rea][1]) for rea in self.result]))
        else:
            name, x, y = list(zip(*[([rea] + self.result[rea][0] + self.result[rea][1])
                                    for rea in self.result]))

        # calculate significance vector
        if self.method in ["fva", "mva"]:
            significance = [1 if abs_diff([xmin[i], xmax[i]],
                                          [ymin[i]*obj_coeff, ymax[i]*obj_coeff])
                            >= style["significance_threshold"]
                            else 0 for i in range(len(name))]
            # draw errorbar (only supported with fva and mva methods)
            if style["show_errorbar"] == "significant":
                plt.errorbar(x=[x[i] for i in range(len(x)) if significance[i] == 1],
                             y=[y[i]
                                 for i in range(len(y)) if significance[i] == 1],
                             xerr=[[xmin[i] for i in range(len(xmin)) if significance[i] == 1],
                                   [xmax[i] for i in range(len(xmax)) if significance[i] == 1]],
                             yerr=[[ymin[i] for i in range(len(ymin)) if significance[i] == 1],
                                   [ymax[i] for i in range(len(ymax)) if significance[i] == 1]],
                             ls='none', elinewidth=1, ecolor="black")
            elif style["show_errorbar"] == "all":
                plt.errorbar(x=x, y=y, xerr=[xmin, xmax], yerr=[ymin, ymax], ls='none',
                             elinewidth=1, ecolor="black")

        else:
            significance = [1 if old_div(abs(x[i]-y[i]*obj_coeff), (old_div((x[i]+y[i]*obj_coeff), 2)))
                            >= style["significance_threshold"] else 0 for i in range(len(name))]

        # add bisecting line
        plt.plot([0, 1000], [0, 1000], ls="--", c=".3")
        # format axis
        plt.ylim(min(y)-min(y)*0.1, (max(y)+max(y)*10))
        plt.xlim(min(x)-min(x)*0.1, (max(x)+max(x)*10))
        plt.yscale("log")
        plt.xscale("log")

        # add title and axis labels
        plt.xlabel(style["xlabel"])
        plt.ylabel(style["ylabel"])
        plt.title(style["title"])

        # add grid to plot (default is False)
        plt.grid(style["plot_grid"])

        # add scatter points
        plt.scatter(x, y, c=[style["significant_color"] if i == 1
                             else style["insignificant_color"] for i in significance])

        # add annotations
        name_abbr = abbreviate_all_in_list(name, max_len=style["label_length"],
                                           ids=style["ids_in_names"], unique=style["unique_names"])

        # use adjustText module if installed:
        if ADJ_TEXT_AV:
            texts = []
            for x_i, y_i, label, sig in zip(x, y, name_abbr, significance):
                if style["show_labels"] == "significant" and sig == 0:
                    continue
                elif style["show_labels"] not in ["all", "significant"]:
                    break
                texts.append(plt.text(x_i, y_i, label))
            if texts:
                adjust_text(texts, color=style["labelcolor"], fontsize=style["fontsize"],
                            arrowprops=dict(arrowstyle="->", color=style["arrowcolor"],
                                            lw=style["arrow_line_width"]))

        # otherwise just do annotations (overlap!)
        else:
            for i, label in enumerate(name_abbr):
                if style["show_labels"] == "significant" and significance[i] == 0:
                    continue
                elif style["show_labels"] not in ["all", "significant"]:
                    break
                plt.annotate(label, fontsize=style["fontsize"], xy=(x[i], y[i]), xytext=(6, 6),
                             color=style["labelcolor"], textcoords='offset points')
        # show interactive plot
        if outputfile:
            try:
                plt.savefig(outputfile, bbox_inches='tight', dpi=300)
                print("Scatterplot was saved as %r." % outputfile)
            except IOError:
                print("Error: No such file or directory: %r" % outputfile)
            except ValueError:
                print ("Error: The format of %r is not supported. "
                       "Please choose one of the following:\n"
                       "eps, jpeg, jpg, pdf, pgf, png, ps, raw, rgba, svg, svgz, tif, tiff"
                       % outputfile)
        else:
            plt.show()


def main():
    """main function"""

    version = "Scatterplot\n" + COPYRIGHT_VERSION_STRING
    parser = argparse.ArgumentParser(version=version)
    # required:
    parser.add_argument("-x", "--xfile", dest="xfile", required=True,
                        help="The input FILE for the x axis. Must be either FBA/FVA/MFM or a "
                        "scenario file for analysis.", metavar="FILE")
    parser.add_argument("-y", "--yfile", dest="yfile", required=True,
                        help="The input FILE for the y axis. Important: Must be of same kind "
                        "as the file for the x axis! ", metavar="FILE")
    parser.add_argument("-m", "--method", required=True,
                        dest="method", help="method for calculation or analysis. Choose "
                        "one of the following: fba, fva, mfm, mva (only with sce files)",
                        metavar="[FBA/FVA/MFM/MVA]")
    # optional:
    parser.add_argument("-r", "--reactions", dest="reactionfile", required=False, default=None,
                        help="Optional: perform Analysis on the network given "
                        "by the reaction FILE", metavar="FILE")
    parser.add_argument("-c", "--config", dest="configfile", required=False, default=None,
                        help="Optional: style file (ini) to define "
                        "the look of the scatterplot", metavar="FILE")
    parser.add_argument("-o", "--output", dest="outputfile", required=False, default=None,
                        help="Optional: write scatterplot to FILE",
                        metavar="FILE")
    parser.add_argument("-b", "--biomass", required=False, default=None,
                        dest="obj_fun", help="Optional: Name of objective function "
                        "to exclude from significance calculation ")

    options = parser.parse_args()

    rea = options.reactionfile
    # if rea:
    #     model = MetabolicModel()
    #     rparser = ReactionParser()
    #     try:
    #         model.addReactionsFromFile(rea, rparser)
    #     except IOError, strerror:
    #         print ("An error occurred while trying to read file %s:" %
    #                os.path.basename(rea))
    #         print strerror
    #         exit()
    #     except SyntaxError, strerror:
    #         print ("Error in reaction file %s:" %
    #                os.path.basename(rea))
    #         print strerror
    #         exit()

    scatter = Scatterplot(method=options.method,
                          filenames=[options.xfile, options.yfile],
                          model=rea if rea else None)

    scatter.do_scatterplot(outputfile=options.outputfile,
                           obj_fun=options.obj_fun,
                           stylefile=options.configfile)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""
This script predicts the directions of reactions based on the upper and lower
bounds for the Gibbs free energy changes computed by getReaEnthalpies(). It
checks whether the given reaction and scenario file conform to the predicted
reaction directions and adds appropriate constraints to the given scenario file
to make reactions irreversible.


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

Michael Stelzer and Rene Rex contributed to this file.

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import absolute_import

from metano.defines import COPYRIGHT_VERSION_STRING
from metano.fba import OptionParser
from metano.optparse import OptionGroup
from metano.reactionparser import ReactionParser
from metano.paramparser import ParamParser
from metano.metabolicmodel import MetabolicModel
from metano.compute_rea_enthalpies import getReaEnthalpies, readReaEnthalpiesFromFile
import os


def main():

    # Parse command line
    usage = ("Usage: %prog <delta-G-file> [options]\n\n"
             "Gibbs free energy changes are taken from the <delta-g-file> if "
             "given.\nElse, they are computed from the data specified via the "
             "-r/-c/-d/-s/-t set of\noptions.")
    version = "%prog\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    standardOptGroup = OptionGroup(parser, "Standard parameters")
    standardOptGroup.add_option("-p", "--parameters", dest="paramFile",
                                help="check the given scenario FILE",
                                metavar="FILE")
    standardOptGroup.add_option("-o", "--output", dest="outputFile",
                                help="write modified scenario file to FILE "
                                "(must not be the same as scenario file)",
                                metavar="FILE")
    standardOptGroup.add_option("-e", "--epsilon", dest="epsilon",
                                help="set THRESHOLD for recognizing Gibbs free "
                                "energy change as clearly positive (default 0)",
                                metavar="THRESHOLD")
    parser.add_option_group(standardOptGroup)

    computeOptGroup = OptionGroup(parser, "Parameters for computation of "
                                  "Gibbs free energies", "These are only needed"
                                  " if delta-G values are not read from file.")
    computeOptGroup.add_option("-r", "--reactions", dest="reactionFile",
                               help="use reactions from reaction FILE",
                               metavar="FILE")
    computeOptGroup.add_option("-c", "--concentrations",
                               dest="concentrationFile", help="use "
                               "concentrations from FILE", metavar="FILE")
    computeOptGroup.add_option("-d", "--thermodynamics", dest="thermodynFile",
                               help="use thermodynamic data from FILE",
                               metavar="FILE")
    computeOptGroup.add_option("-s", "--synonyms", dest="synonymFile",
                               help="use metabolite synonyms from FILE",
                               metavar="FILE")
    computeOptGroup.add_option("-t", "--temperature", dest="temperature",
                               help="set temperature to VALUE (in degrees "
                               "Celsius)", metavar="VALUE")
    parser.add_option_group(computeOptGroup)
    parser.set_defaults(epsilon='0.')

    options, args = parser.parse_args()
    parser.check_required("-p")
    parser.check_required("-o")
    parser.check_required("-t")

    try:
        epsilon = float(options.epsilon)
    except ValueError:
        print ("Error: Invalid floating point value for epsilon (%s)" %
               options.epsilon)
        exit()
    if (os.path.exists(options.outputFile) and
            os.path.samefile(options.outputFile, options.paramFile)):
        print ("Error: Input and output scenario files are the same (%s)" %
               options.paramFile)
        exit()

    if epsilon < 0.:
        print("Warning: epsilon < 0. Using default value of 0 instead.")
        epsilon = 0.

    if len(args) > 0:
        gibbsR = readReaEnthalpiesFromFile(args[0])
    else:
        print ("\nInfo: No file with Gibbs free energies given. Launching "
               "computation of values.\n")
        parser.check_required("-r")
        parser.check_required("-c")
        parser.check_required("-d")
        parser.check_required("-s")
        parser.check_required("-t")
        try:
            temperature = float(options.temperature)
        except ValueError:
            print ("Error: Invalid floating point value for temperature (%s)" %
                   options.temperature)
            exit()

        # Compute Gibbs free energies from the given data
        gibbsR = getReaEnthalpies(options.concentrationFile,
                                  options.thermodynFile, options.synonymFile,
                                  options.reactionFile, temperature)

    # Parse scenario file
    pparser = ParamParser()
    try:
        # Parse file
        maxmin, obj_name, solver, numiter, lb, ub =\
            pparser.parse(options.paramFile)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.paramFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in scenario file %s:" %
               os.path.basename(options.paramFile))
        print(strerror)
        exit()
    except ValueError as strerror:
        print(strerror)
        exit()

    if options.reactionFile:
        # Parse reaction file
        rparser = ReactionParser()
        model = MetabolicModel()
        try:
            model.addReactionsFromFile(options.reactionFile, rparser)
        except IOError as strerror:
            print ("An error occurred while trying to read file %s:" %
                   os.path.basename(options.reactionFile))
            print(strerror)
            exit()
        except SyntaxError as strerror:
            print ("Error in reaction file %s:" %
                   os.path.basename(options.reactionFile))
            print(strerror)
            exit()

        # Set flux bounds in model
        model_messages = []
        try:
            model.setFiniteBounds(lb, ub, True, model_messages)
        except ValueError as strerror:
            print(strerror)
            exit()
        reactions = model.getReactionNames()
        # Include implicit flux bounds from reaction file
        p_lb, p_ub = lb, ub
        lb, ub = model.getFiniteBounds()

        # Show warning and info messages of parsers
        msgs = (rparser.getMessages() + pparser.getMessages() +
                [x[1] for x in model_messages])
        if msgs:
            print('\n'+'\n'.join(msgs))
    else:
        print ("Warning: No reaction file given. Implicit bounds defined by "
               "reaction\nreversibilities cannot be taken into account.")
        p_lb, p_ub = lb, ub
        reactions = sorted(
            set(list(gibbsR.keys())+list(lb.keys())+list(ub.keys())))

    # Copy optimization parameters
    param_head = ("OBJ %s %s\nSOLVER %s\n" % ({True: "MAX", False: "MIN"}[maxmin],
                                              obj_name, solver))
    if numiter >= 0:
        param_head += "NUM_ITER %u\n" % numiter
    param_head += '\n'

    try:
        with open(options.outputFile, 'w') as f:
            f.write(param_head)
            for rea in reactions:
                if rea in gibbsR:
                    mini, maxi = gibbsR[rea]
                    if maxi < -epsilon:
                        # Reaction is irreversible and proceeds from
                        # left to right (LB 0)
                        if rea in ub and ub[rea] <= 0.:
                            print ("Warning: Gibbs free energy predicts "
                                   "LB %s 0,\nbut reaction is defined as UB %g "
                                   "(skipped)" % (rea, ub[rea]))
                            if rea in p_lb:
                                f.write("LB %s %g\n" % (rea, p_lb[rea]))
                            if rea in p_ub:
                                f.write("UB %s %g # Conflict (delta_G_max = %g)"
                                        "\n" % (rea, p_ub[rea], maxi))
                        elif rea in lb and lb[rea] >= 0.:
                            if rea in p_lb:
                                f.write("LB %s %g # Direction confirmed by "
                                        "delta_G_max = %g\n" % (rea, p_lb[rea],
                                                                maxi))
                            if rea in p_ub:
                                f.write("UB %s %g\n" % (rea, p_ub[rea]))
                        else:
                            # (rea not in ub or ub[rea] > 0) and (rea not in lb
                            # or lb[rea] < 0)
                            # Add LB 0 line
                            f.write("LB %s %g # Assigned because delta_G_max = "
                                    "%g\n" % (rea, 0., maxi))
                            if rea in p_ub:
                                f.write("UB %s %g\n" % (rea, p_ub[rea]))
                    elif mini > epsilon:
                        # Reaction is irreversible and proceeds from
                        # right to left (UB 0)
                        if rea in lb and lb[rea] >= 0.:
                            print ("Warning: Gibbs free energy predicts "
                                   "UB %s 0,\nbut reaction is defined as LB %g "
                                   "(skipped)" % (rea, lb[rea]))
                            if rea in p_lb:
                                f.write("LB %s %g # Conflict (delta_G_min = %g)"
                                        "\n" % (rea, p_lb[rea], mini))
                            if rea in p_ub:
                                f.write("UB %s %g\n" % (rea, p_ub[rea]))
                        elif rea in ub and ub[rea] <= 0.:
                            if rea in p_lb:
                                f.write("LB %s %g\n" % (rea, p_lb[rea]))
                            if rea in p_ub:
                                f.write("UB %s %g # Direction confirmed by "
                                        "delta_G_min = %g\n" % (rea, p_lb[rea],
                                                                mini))
                        else:
                            # (rea not in ub or ub[rea] > 0) and (rea not in lb
                            # or lb[rea] < 0)
                            if rea in p_lb:
                                f.write("LB %s %g\n" % (rea, p_lb[rea]))
                            # Add UB 0 line
                            f.write("UB %s %g # Assigned because delta_G_min = "
                                    "%g\n" % (rea, 0., mini))
                    else:
                        # Just copy parameters
                        if rea in p_lb:
                            f.write("LB %s %g\n" % (rea, p_lb[rea]))
                        if rea in p_ub:
                            f.write("UB %s %g\n" % (rea, p_ub[rea]))
                else:
                    # Just copy parameters
                    if rea in p_lb:
                        f.write("LB %s %g\n" % (rea, p_lb[rea]))
                    if rea in p_ub:
                        f.write("UB %s %g\n" % (rea, p_ub[rea]))
    except IOError as strerror:
        print ("Unable to write to file %s:" %
               os.path.basename(options.outputFile))
        print(strerror)
        exit()


if __name__ == "__main__":
    main()
